import logging
from typing import Union, Optional, Dict, List, Callable, Tuple

import matplotlib
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import sklearn
from plotly import colors
from plotly.subplots import make_subplots
from sklearn import model_selection
from sklearn.base import is_classifier, is_regressor
from sklearn.model_selection import cross_validate
from sklearn.pipeline import Pipeline

from elphick.sklearn_viz.model_selection.metrics import regression_metrics, classification_metrics
from elphick.sklearn_viz.model_selection.scorers import classification_scorers, regression_scorers


def subplot_index(idx: int, col_wrap: int) -> Tuple[int, int]:
    col: int = int(idx % col_wrap + 1)
    row: int = int(np.floor(idx / col_wrap) + 1)
    return row, col


def plot_model_selection(algorithms: Union[sklearn.base.RegressorMixin, sklearn.base.ClassifierMixin, Dict],
                         datasets: Union[pd.DataFrame, Dict],
                         target: str,
                         pre_processor: Optional[Pipeline] = None,
                         k_folds: int = 10,
                         title: Optional[str] = None) -> go.Figure:
    """

    Args:
            algorithms: sklearn estimator or a Dict of algorithms to cross-validate, keyed by string name/code.
            datasets: pandas DataFrame or a dict of DataFrames, keyed by string name/code.
            target: target column
            pre_processor: Optional pipeline used to pre-process the datasets.
            k_folds: The number of cross validation folds.
            title: Optional plot title

    Returns:
        a plotly GraphObjects.Figure

    """

    return ModelSelection(algorithms=algorithms, datasets=datasets, target=target, pre_processor=pre_processor,
                          k_folds=k_folds).plot(title=title)


class ModelSelection:
    def __init__(self,
                 algorithms: Union[
                     sklearn.base.RegressorMixin, sklearn.base.ClassifierMixin, Dict[str, sklearn.base.BaseEstimator]],
                 datasets: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 target: str,
                 pre_processor: Optional[Pipeline] = None,
                 k_folds: int = 10,
                 scorer: Optional[Union[str, Callable]] = None,
                 metrics: Optional[Dict[str, Callable]] = None,
                 group: Optional[pd.Series] = None,
                 random_state: Optional[int] = None):
        """

        Args:
            algorithms: sklearn estimator or a Dict of algorithms to cross-validate, keyed by string name/code.
            datasets: pandas DataFrame or a dict of DataFrames, keyed by string name/code.
            target: target column
            pre_processor: Optional pipeline used to pre-process the datasets.
            k_folds: The number of cross validation folds.
            scorer: Optional callable scorers which the model will be fitted using
            metrics: Optional Dict of callable metrics to calculate post-fitting
            group: Optional group variable by which to partition/group metrics.  The same group applies across all
             datasets, so is more useful when testing different algorithms.
            random_state: Optional random seed
        """
        self._logger = logging.getLogger(name=__class__.__name__)
        self.pre_processor: Pipeline = pre_processor
        if isinstance(algorithms, sklearn.base.BaseEstimator):
            self.algorithms = {algorithms.__class__.__name__: algorithms}
        else:
            self.algorithms = algorithms
        if isinstance(datasets, pd.DataFrame):
            self.datasets = {'Dataset': datasets}
        else:
            self.datasets = datasets
        self.target = target
        self.k_folds: int = k_folds

        self.is_classifier: bool = is_classifier(list(self.algorithms.values())[0])
        self.is_regressor: bool = is_regressor(list(self.algorithms.values())[0])
        if scorer is not None:
            self.scorer = scorer
        else:
            self.scorer = classification_scorers[list(classification_scorers.keys())[0]] if self.is_classifier else \
                regression_scorers[list(regression_scorers.keys())[0]]

        if metrics is not None:
            self.metrics = metrics
        else:
            self.metrics = classification_metrics if self.is_classifier else regression_metrics

        self.group: pd.Series = group
        self.random_state: Optional[int] = random_state

        self.features_in: List[str] = [col for col in self.datasets[list(self.datasets.keys())[0]] if
                                       col != self.target]

        self._data: Optional[Dict] = None
        self._num_algorithms: int = len(list(self.algorithms.keys()))
        self._num_datasets: int = len(list(self.datasets.keys()))

        if self._num_algorithms > 1 and self._num_datasets > 1:
            raise NotImplementedError("Cannot have multiple algorithms and multiple datasets.")

    @property
    def data(self) -> Optional[Dict]:
        if self.metrics is None:
            cv_kwargs: Dict = dict()
        else:
            cv_kwargs: Dict = dict(return_estimator=True, return_indices=True, error_score=np.nan)

        if self._data is not None:
            results = self._data
        else:
            results: Dict = {}
            for data_key, data in self.datasets.items():
                self._logger.info(f"Commencing Cross Validation for dataset {data_key}")
                results[data_key] = {}
                x: pd.DataFrame = data[self.features_in]
                y: pd.DataFrame = data[self.target]
                if self.pre_processor:
                    x = self.pre_processor.set_output(transform="pandas").fit_transform(X=x)

                for algo_key, algo in self.algorithms.items():
                    kfold = model_selection.KFold(n_splits=self.k_folds, random_state=self.random_state, shuffle=True)
                    res = cross_validate(algo, x, y, cv=kfold, scoring=self.scorer, **cv_kwargs)
                    if self.metrics is not None:
                        res['metrics'], res['metrics_group'] = self.calculate_metrics(x=x, y=y,
                                                                                      estimators=res['estimator'],
                                                                                      indices=res['indices'],
                                                                                      group=self.group)
                    results[data_key][algo_key] = res
                    res_mean = res[f"test_score"].mean()
                    res_std = res[f"test_score"].std()
                    self._logger.info(f"CV Results for {algo_key}: Mean = {res_mean}, SD = {res_std}")

            self._data = results

        return results

    def plot(self,
             metrics: Optional[Union[str, List[str]]] = None,
             show_group: bool = False,
             title: Optional[str] = None,
             col_wrap: Optional[int] = None) -> go.Figure:
        """Create the plot

        The plot will show the cross-validation scores for each algorithm and dataset.  The first panel is used to show
        the scorer, that is the metric used to fit the model.  If multiple metrics are supplied, each metric will be
        shown in a separate panel.  If a show_group is true, the metrics will be grouped by the group variable.
        col_wrap allows the width of the plot to be controlled by wrapping the columns to new rows.

        KUDOS: https://towardsdatascience.com/applying-a-custom-colormap-with-plotly-boxplots-5d3acf59e193

        Args:
            metrics: The metric or metrics to plot in addition to the scorer.  Each metric will be plotted in a
             separate panel.
            show_group: If True (and a group variable has been set), plot by group.
            title: Title of the plot
            col_wrap: If plotting multiple metrics, col_wrap will wrap columns to new rows, resulting in
             col-wrap columns, and multiple rows.

        Returns:
            a plotly GraphObjects.Figure

        """

        data: pd.DataFrame = self.get_cv_scores()
        data = data.droplevel(level=0, axis=1) if self._num_datasets == 1 else data.droplevel(level=1, axis=1)

        metric_data: pd.DataFrame = pd.DataFrame()

        if metrics is not None:
            if isinstance(metrics, str):
                metrics = [metrics]
            metric_data = self.get_cv_metrics(metrics, show_group)
        else:
            metrics = []

        if self._num_algorithms > 1:
            x_index = 'algo_key'
        else:
            x_index = 'data_key'

        # define the color map for the scorer
        vmin, vmax = data.min().min(), data.max().max()
        norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
        cmap = matplotlib.cm.get_cmap('RdYlGn')

        subtitle: str = f'Cross Validation folds={self.k_folds}'
        if title is None:
            title = subtitle
        else:
            title = title + '<br>' + subtitle

        # create the plot, managing the shape.
        num_plots: int = len(metrics) + 1 if len(metrics) > 0 else 1
        num_cols: int = num_plots if col_wrap is None else col_wrap
        num_rows, _ = subplot_index(len(metrics), col_wrap=num_cols)
        fig = make_subplots(rows=num_rows, cols=num_cols, subplot_titles=[f'{self.scorer} (scorer)'] + metrics)

        # Add the scorer subplot
        for col in data.columns:
            # For the scorer build the plot by column to color individually based on score
            median = np.median(data[col])  # find the median
            color = 'rgb' + str(cmap(norm(median))[0:3])  # normalize
            fig.add_trace(go.Box(y=data[col], name=col, boxpoints='all', notched=True, fillcolor=color,
                                 line={"color": "grey"}, marker={"color": "grey"}, showlegend=False,
                                 offsetgroup='A'), row=1, col=1)

        # add the metric subplots
        for i, metric in enumerate(metrics):
            row, col = subplot_index(i + 1, col_wrap=num_cols)
            if show_group:
                colorscale = colors.qualitative.Plotly + colors.qualitative.Dark24
                add_to_legend = True if i == 0 else False
                df_metric: pd.DataFrame = metric_data.query('metric==@metric').drop(columns=['metric'])
                x = df_metric.index.get_level_values(x_index)
                for g, grp in enumerate(df_metric.columns):
                    if len(df_metric.columns) > len(colorscale):
                        raise ValueError("Too many groups to plot")
                    fig.add_trace(go.Box(x=x, y=df_metric[grp], name=grp, boxpoints='all', notched=True,
                                         legendgroup=self.group.name,
                                         showlegend=add_to_legend,
                                         line={"color": colorscale[g]}, marker={"color": colorscale[g]},
                                         offsetgroup=str(g)), row=row, col=col)
            else:
                df_metric: pd.DataFrame = metric_data.query('metric==@metric').drop(columns=['metric'])
                x = list(df_metric.index.get_level_values(x_index))
                fig.add_trace(go.Box(x=x, y=df_metric.values.ravel(), name=metric, boxpoints='all', notched=True,
                                     line={"color": "grey"}, marker={"color": "grey"}), row=row, col=col)

        # finalise some display elements
        fig.update_layout(title=title, showlegend=False)
        if show_group:
            fig.update_layout(boxmode='group', showlegend=True, legend_title=self.group.name,
                              boxgroupgap=0.5, boxgap=0
                              )

        return fig

    def plot_category_analysis(self,
                               algorithm: Optional[str] = None,
                               dataset: Optional[str] = None,
                               metrics: Optional[Union[str, List[str]]] = None,
                               title: Optional[str] = None,
                               col_wrap: Optional[int] = None) -> go.Figure:
        """Plot the category feature analysis

        Args:
            algorithm: If supplied, this will be the name of the algorithm tested.  If None the first algorithm is used.
            dataset: If supplied, this will be the name of the dataset tested.  If None the first dataset is used.
            metrics: The metric or metrics to show.  Each metric will be plotted in a
             separate panel.
            title: Title of the plot
            col_wrap: If plotting multiple metrics, col_wrap will wrap columns to new rows, resulting in
             col-wrap columns, and multiple rows.

        Returns:
            a plotly GraphObjects.Figure

        """
        algorithms: list[str] = list(self.algorithms.keys())
        algorithm: str = algorithms[0] if algorithm is None else algorithm
        if algorithm not in algorithms:
            raise KeyError(f"Algorithm {algorithm} is not in the list of available algorithms: {algorithms}")

        datasets: list[str] = list(self.datasets.keys())
        dataset: str = datasets[0] if dataset is None else dataset
        if dataset not in datasets:
            raise KeyError(f"Dataset {dataset} is not in the list of available datasets: {datasets}")

        metrics: list[str] = [list(self.metrics.keys())[0]] if metrics is None else metrics

        baseline_metrics: pd.DataFrame = self.get_cv_metrics(metrics, by_group=True).loc[
            (slice(None), dataset, algorithm)]
        baseline_metrics = baseline_metrics.melt(id_vars=['metric'], value_vars=self.group.unique().tolist(),
                                                 var_name='group',
                                                 ignore_index=False).assign(model='baseline')

        # cross-validate the individual models
        by_group_metrics, by_group_scores = self.get_model_by_group_data(algorithm, dataset)
        by_group_metrics = by_group_metrics.melt(id_vars=['group'], value_vars=metrics, var_name='metric',
                                                 ignore_index=False).assign(model='by_group')
        metric_data: pd.DataFrame = pd.concat([baseline_metrics, by_group_metrics]).sort_values(['model', 'metric'])
        metric_data = metric_data.set_index(['metric', 'group'], append=True).pivot(columns='model',
                                                                                    values='value').reset_index(
            'metric')

        if title is None:
            title = f'Model by Group Test on {algorithm} with {self.k_folds} folds'

        num_plots: int = len(metrics) if len(metrics) > 0 else 1
        num_cols: int = num_plots if col_wrap is None else col_wrap
        num_rows, _ = subplot_index(len(metrics) - 1, col_wrap=num_cols)
        fig = make_subplots(rows=num_rows, cols=num_cols,
                            subplot_titles=metrics)

        # metrics
        for i, metric in enumerate(metrics):
            row, col = subplot_index(i, col_wrap=num_cols)
            colorscale = colors.qualitative.Plotly
            add_to_legend = True if i == 0 else False
            df_metric: pd.DataFrame = metric_data.query('metric==@metric').drop(columns=['metric'])
            x = df_metric.index.get_level_values('group')
            for g, grp in enumerate(df_metric.columns):
                fig.add_trace(go.Box(x=x, y=df_metric[grp], name=grp, boxpoints='all', notched=True,
                                     legendgroup=self.group.name,
                                     showlegend=add_to_legend,
                                     line={"color": colorscale[g]}, marker={"color": colorscale[g]},
                                     offsetgroup=str(g)), row=row, col=col)

        fig.update_layout(title=title, showlegend=False)
        fig.update_layout(boxmode='group', showlegend=True, legend_title='model',
                          boxgroupgap=0.5, boxgap=0
                          )

        return fig

    def get_model_by_group_data(self, algorithm, dataset):
        results: dict = {}
        by_group_score_chunks: list = []
        by_group_metric_chunks: list = []
        for grp in self.group.unique():
            grp_index: pd.Index = self.group.loc[self.group == grp].index
            x: pd.DataFrame = self.datasets[dataset][self.features_in].loc[grp_index]
            y: pd.DataFrame = self.datasets[dataset][self.target].loc[grp_index]
            if self.pre_processor:
                x = self.pre_processor.set_output(transform="pandas").fit_transform(X=x)
            kfold = model_selection.KFold(n_splits=self.k_folds, random_state=self.random_state, shuffle=True)
            res = cross_validate(self.algorithms[algorithm], x, y, cv=kfold, scoring=self.scorer, return_estimator=True,
                                 return_indices=True)
            by_group_score_chunks.append(pd.Series(res['test_score'], name=grp))
            if self.metrics is not None:
                res['metrics'], _ = self.calculate_metrics(x=x, y=y,
                                                           estimators=res['estimator'],
                                                           indices=res['indices'],
                                                           group=None)
            results[grp] = res
            by_group_metric_chunks.append(
                pd.DataFrame(res['metrics']).assign(group=grp).rename_axis('fold', axis='index'))
        by_group_metrics: pd.DataFrame = pd.concat(by_group_metric_chunks)
        #                                   .reset_index().melt(id_vars=['fold', 'group'],
        #                                                                                       value_vars=list(
        #                                                                                           self.metrics.keys()),
        #                                                                                       var_name='metric'))
        # by_group_metrics = by_group_metrics.set_index(['fold', 'metric', 'group']).unstack(level=-1)
        # by_group_metrics = by_group_metrics.droplevel(level=0, axis=1).reset_index(-1)
        by_group_scores = pd.concat(by_group_score_chunks, axis=1)

        return by_group_metrics, by_group_scores

    def get_cv_scores(self) -> pd.DataFrame:
        chunks: List = []
        for data_key, data in self.datasets.items():
            for algo_key, algo in self.algorithms.items():
                chunks.append(pd.Series(self.data[data_key][algo_key][f"test_score"], name=(data_key, algo_key)))
        return pd.concat(chunks, axis=1)

    def get_cv_metrics(self, metrics, by_group: bool = False) -> pd.DataFrame:
        chunks: List = []
        metric_key = "metrics_group" if by_group else "metrics"
        for data_key, data in self.datasets.items():
            for algo_key, algo in self.algorithms.items():
                for metric in metrics:
                    chunks.append(pd.DataFrame(self.data[data_key][algo_key][metric_key][metric]).assign(
                        **dict(data_key=data_key, algo_key=algo_key, metric=metric)))
        res: pd.DataFrame = pd.concat(chunks, axis=0).set_index(['data_key', 'algo_key'], append=True).rename(
            columns={0: 'value'})
        res.index.names = ['fold', 'data_key', 'algo_key']
        return res

    def calculate_metrics(self, x, y, estimators, indices, group) -> Tuple[Dict, Dict]:
        metric_results: Dict = {}
        metric_results_group: Dict = {}

        for k, fn_metric in self.metrics.items():
            metric_values: List = []
            metric_groups: Dict = {}
            for estimator, test_indexes in zip(estimators, indices['test']):
                y_true = y[y.index[test_indexes]]
                y_est = estimator.predict(x.loc[x.index[test_indexes], :])
                if isinstance(y_est, pd.DataFrame) and y_est.shape[1] == 1:
                    y_est = y_est.iloc[:, 0]
                metric_values.append(fn_metric(y_true, y_est))
                if group is not None:
                    # calculate the metric by each group in the group series.
                    y_est = pd.merge(left=pd.Series(y_est, name='y_est', index=x.index[test_indexes]),
                                     right=group, left_index=True, right_index=True)
                    y_est_grouped = y_est.groupby([group.name])
                    grouped_results = [y_est_grouped.get_group(x) for x in y_est_grouped.groups]
                    for grp_res in grouped_results:
                        group_value = str(grp_res[group.name].iloc[0])
                        group_metric_results = fn_metric(y_true[grp_res.index], grp_res['y_est'].values)
                        if group_value not in metric_groups.keys():
                            metric_groups[group_value] = [group_metric_results]
                        else:
                            metric_groups[group_value].append(group_metric_results)
            metric_results[k] = metric_values
            if group is not None:
                metric_results_group[k] = metric_groups

        return metric_results, metric_results_group

# if __name__ == '__main__':
#
#     for i in range(0, 7):
#         print(subplot_index(i, col_wrap=3))
#
#     print('done')



import logging
from typing import Optional, Union

import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted
import plotly.graph_objects as go
import plotly.express as px


class Errors:
    def __init__(self,
                 mdl,
                 x_test: Optional[pd.DataFrame] = None,
                 y_test: Optional[Union[pd.DataFrame, pd.Series]] = None):
        """

        Args:
            mdl: The scikit-learn model or pipeline.
            x_test: X values provided to calculate residuals.
            y_test: y values provided to calculate residuals.
        """
        self._logger = logging.getLogger(name=__class__.__name__)
        self.mdl = mdl
        self.X_test: Optional[pd.DataFrame] = x_test
        self.y_test: Optional[Union[pd.DataFrame, pd.Series]] = y_test
        self._data: Optional[pd.DataFrame] = None
        self.is_pipeline: bool = isinstance(mdl, Pipeline)

        check_is_fitted(mdl[-1]) if self.is_pipeline else check_is_fitted(mdl)

    def plot(self, color: Optional[str] = None, title: Optional[str] = 'Error Plot') -> go.Figure:
        """

        Args:
            color: The variable name to color (group) by.
            title: title for the plot

        Returns:
            a plotly GraphObjects.Figure

        """
        y_est = pd.Series(self.mdl.predict(self.X_test), name=f"{self.y_test.name}_est", index=self.X_test.index)
        y = self.y_test
        data = pd.concat([y, y_est], axis='columns')
        lims = [data.min().min(), data.max().max()]
        fig = px.scatter(
            data,
            x=y.name, y=y_est.name,
            marginal_x='histogram', marginal_y='histogram',
            color=color, trendline='ols'
        )
        fig.update_traces(histnorm='probability', selector={'type': 'histogram'})
        fig.add_shape(
            type="line", line=dict(dash='dash'),
            x0=lims[0], y0=lims[0],
            x1=lims[1], y1=lims[1]
        )

        fig.update_layout(title=title)

        return fig



"""
Developed from the example here: https://plotly.com/python/pca-visualization/
"""
import dataclasses
import logging
from typing import Optional, List, Dict

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
from pandas.core.dtypes.common import is_numeric_dtype
from sklearn.decomposition import PCA
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from elphick.sklearn_viz.utils import log_timer


def plot_principal_components(x: pd.DataFrame,
                              color: Optional[pd.Series] = None,
                              plot_3d: bool = True,
                              loading_vectors: bool = True,
                              standardised: bool = False,
                              title: Optional[str] = None) -> go.Figure:
    """

    Args:
        x: X values to transform and plot.
        color: optional series by which to color the markers
        plot_3d: If True plot the top 3 principal components in 3D, otherwise the top 2 in 2D.
        loading_vectors: If True and plot_type is '2D'|'3D' loading vectors will be displayed.
        standardised: If True, plot the standardised PCA, where vectors are transformed to zero mean and
             unit variance.
        title: Optional plot title

    Returns:
        a plotly GraphObjects.Figure

    """

    return PrincipalComponents(x=x, color=color).plot_principal_components(plot_3d=plot_3d,
                                                                           loading_vectors=loading_vectors,
                                                                           standardised=standardised,
                                                                           title=title)


def plot_explained_variance(x: pd.DataFrame, y: Optional[pd.Series] = None,
                            title: Optional[str] = None) -> go.Figure:
    """Plot the cumulative explained variance by principal component.

    Args:
        x: X values to transform and plot.
        y: optional target vector
        title: Optional plot title

    Returns:

    """
    return PrincipalComponents(x=x, color=y).plot_explained_variance(title=title)


def plot_scatter_matrix(x: pd.DataFrame, y: Optional[pd.Series] = None,
                        original_features: bool = False, title: Optional[str] = None) -> go.Figure:
    """Plot a scatter matrix

    Args:
        x: X values to transform and plot.
        y: optional series by which to color the markers
        original_features: If True, plot the original features, otherwise plot the principal components.
        title: Optional plot title

    Returns:

    """
    return PrincipalComponents(x=x, color=y).plot_scatter_matrix(original_features=original_features, title=title)


def plot_loading_vectors(x: pd.DataFrame,
                         color: Optional[pd.Series] = None,
                         standardised: bool = False,
                         title: Optional[str] = None) -> go.Figure:
    """

    Args:
        x: X values to transform and plot.
        color: optional series by which to color the markers
        standardised: If True, plot the standardised PCA, where vectors are transformed to zero mean and
             unit variance.
        title: Optional plot title

    Returns:
        a plotly GraphObjects.Figure

    """

    return PrincipalComponents(x=x, color=color).plot_loading_vectors(standardised=standardised, title=title,
                                                                      by_color=color is not None)


def plot_correlation_circle(x: pd.DataFrame,
                            color: Optional[pd.Series] = None,
                            title: Optional[str] = None) -> go.Figure:
    """

    Args:
        x: X values to transform and plot.
        color: optional series by which to color the markers
        title: Optional plot title

    Returns:
        a plotly GraphObjects.Figure

    """

    return PrincipalComponents(x=x, color=color).plot_loading_vectors(standardised=True, title=title,
                                                                      by_color=color is not None)


@dataclasses.dataclass
class PCResults:
    """Class to hold Principal Component results"""
    data: pd.DataFrame
    explained_variance: pd.Series
    loadings: pd.DataFrame


class PrincipalComponents:
    def __init__(self, x: pd.DataFrame, color: Optional[pd.Series] = None):
        """

        Args:
            x: X values to transform and plot.
            color: the optional series by which to color the markers
        """
        self._logger = logging.getLogger(name=__class__.__name__)
        self.x: pd.DataFrame = x
        self.color: Optional[pd.Series] = color

        self._data: Optional[Dict] = None

    @property
    @log_timer
    def data(self) -> Optional[Dict]:
        def get_pca_results(pipe, x):
            xt: pd.DataFrame = pipe.fit_transform(x)
            xt.columns = [f"PC{i + 1}" for i in range(len(x.columns))]
            var: pd.Series = pd.Series(data=pipe['pca'].explained_variance_ratio_ * 100., name='explained_variance')
            loadings = pd.DataFrame(data=pipe['pca'].components_.T * np.sqrt(pipe['pca'].explained_variance_),
                                    index=x.columns, columns=xt.columns)
            return PCResults(data=xt, explained_variance=var, loadings=loadings)

        if self._data is not None:
            res = self._data
        else:
            res: Dict = {}
            self._logger.info("Commencing PCA")
            pca = make_pipeline(PCA()).set_output(transform="pandas")
            pca_std = make_pipeline(StandardScaler(), PCA()).set_output(transform="pandas")
            for label, pipe in {'raw': pca, 'std': pca_std}.items():
                res[label] = get_pca_results(pipe=pipe, x=self.x)
                if (self.color is not None) & (not is_numeric_dtype(self.color)):
                    for grp in self.color.unique():
                        if 'group' not in res.keys():
                            res['group'] = dict()
                        if grp not in res['group'].keys():
                            res['group'][grp] = dict()
                        res['group'][grp][label] = get_pca_results(pipe=pipe, x=self.x.loc[self.color == grp, :])
            self._data = res

        return res

    def plot_principal_components(self,
                                  plot_3d: bool = False,
                                  loading_vectors: bool = True,
                                  standardised: bool = False,
                                  title: Optional[str] = None) -> go.Figure:
        """Create the pca plot

        Args:
            plot_3d: If True plot the top 3 principal components in 3D, otherwise the top 2 in 2D.
            loading_vectors: If True and plot_type is '2D'|'3D' loading vectors will be displayed.
            standardised: If True, plot the standardised PCA, where vectors are transformed to zero mean and
             unit variance.
            title: Optional plot title

        Loading vectors are implemented manually rather than with annotations (lines with arrows),
         the problem is described well here:
         https://community.plotly.com/t/set-pca-loadings-aka-arrows-in-a-3d-scatter-plot/72905

        Returns:
            a plotly GraphObjects.Figure

        """
        label: str = 'std' if standardised else 'raw'
        pca_data: pd.DataFrame = self.data[label].data
        pca_loadings: pd.DataFrame = self.data[label].loadings
        pca_variance: pd.DataFrame = self.data[label].explained_variance

        df_plot: pd.DataFrame = pd.concat([pca_data, self.x], axis=1).reset_index()
        if plot_3d:
            fig = px.scatter_3d(df_plot, x='PC1', y='PC2', z='PC3',
                                color=self.color, hover_data=list(self.x.reset_index().columns))
            fig.update_traces(marker_size=4)
            if loading_vectors:

                annots: List = [dict(x=row.PC1, y=row.PC2, z=row.PC3,
                                     text=i, showarrow=False,
                                     xanchor="left", xshift=10, yshift=10, opacity=0.7) for i, row in
                                pca_loadings.iterrows()]
                fig.update_layout(scene=dict(annotations=annots))
                for feature_name, row in pca_loadings.iterrows():
                    # noinspection PyTypeChecker
                    fig.add_trace(
                        go.Scatter3d(x=(row.PC1,), y=(row.PC2,), z=(row.PC3,), mode='markers',
                                     marker={'size': 6, 'line': dict(width=2, color='black')},
                                     name=feature_name,
                                     showlegend=True,
                                     legendgroup="features",
                                     legendgrouptitle_text="feature vectors",
                                     ))
                    fig.add_trace(
                        go.Scatter3d(x=(0, row.PC1), y=(0, row.PC2), z=(0, row.PC3), mode='lines',
                                     line={'width': 5, 'color': 'black'},
                                     name=feature_name,
                                     showlegend=False))
                fig.update_layout(legend=dict(groupclick="toggleitem"))
                title = (f"Top 3 Principal Components<br>Explained Variance = "
                         f"{round(pca_variance.iloc[0:3].sum(), 1)}%") if title is None else title
        else:  # 2D
            fig = px.scatter(df_plot, x='PC1', y='PC2',
                             color=self.color, hover_data=list(self.x.reset_index().columns))
            fig.update_traces(marker_size=5)

            if loading_vectors:
                loadings = pca_loadings.iloc[:, 0:2]
                self.add_loading_vectors(fig, loadings)
            title = (f"Top 2 Principal Components<br>Explained Variance = "
                     f"{round(pca_variance.iloc[0:2].sum(), 1)}%") if title is None else title

        fig.update_layout(legend_title_text=self.color.name)
        fig.update_layout(title=title,
                          xaxis_title=f"PC1 ({round(self.data['std'].explained_variance.iloc[0], 1)}%)",
                          yaxis_title=f"PC2 ({round(self.data['std'].explained_variance.iloc[1], 1)}%)")
        if self.color is not None:
            fig.update_layout(coloraxis_colorbar_title_text=self.color.name)

        return fig

    def plot_explained_variance(self,
                                standardised: bool = False,
                                title: Optional[str] = None) -> go.Figure:
        """Plot the cumulative explained variance by principal component.

        Args:
            standardised: If True, plot the standardised PCA, where vectors are transformed to zero mean and
             unit variance.
            title: Optional plot title

        Returns:

        """
        pca_variance: pd.DataFrame = self.data['std'].explained_variance if standardised else self.data[
            'raw'].explained_variance
        exp_var_cumul = np.cumsum(pca_variance)
        fig = px.area(
            x=range(1, exp_var_cumul.shape[0] + 1),
            y=exp_var_cumul,
            labels={"x": "# Components", "y": "Explained Variance"}
        )
        title = 'Cumulative Explained Variance by Principal Component' if title is None else title
        fig.update_layout(title=title)
        fig.update_xaxes(type='category')

        return fig

    def plot_scatter_matrix(self, original_features: bool = False, standardised: bool = False,
                            title: Optional[str] = None) -> go.Figure:
        """Plot a scatter matrix

        Args:
            original_features: If True, plot the original features, otherwise plot the principal components.
            standardised: If True, plot the standardised PCA, where vectors are transformed to zero mean and
             unit variance.
            title: Optional plot title

        Returns:

        """
        label: str = 'std' if standardised else 'raw'
        y = self.color
        if original_features:
            x = self.x
            title = 'Scatter Matrix - Original Feature Space' if title is None else title
        else:
            x = self.data[label].data
            title = 'Scatter Matrix - All Principal Components' if title is None else title

        if original_features:
            df_plot: pd.DataFrame = pd.concat([x, y], axis=1).reset_index()
            hover_data = ['index' if x.index.name is None else x.index.name]
        else:
            df_plot: pd.DataFrame = pd.concat([x, y, self.x], axis=1).reset_index()
            hover_data = list(self.x.reset_index().columns)

        fig = px.scatter_matrix(data_frame=df_plot, dimensions=list(x.columns),
                                color=y.name, hover_data=hover_data)
        fig.update_traces(diagonal_visible=False)
        title = 'Top 3 Principal Components' if title is None else title
        fig.update_layout(title=title)

        return fig

    def plot_loading_vectors(self, standardised: bool = False, by_color: bool = False,
                             title: Optional[str] = None) -> go.Figure:
        """plot the loading vectors.

        Args:
            standardised: If True, plot the standardised PCA, where vectors are transformed to zero mean and
             unit variance.
             by_color: If True, plot the loading vectors by color group.
            title: Optional plot title

        Returns:
            a plotly GraphObjects.Figure

        """
        label: str = 'std' if standardised else 'raw'
        if by_color:
            chunks = []

            for grp, d_label in self.data['group'].items():
                chunks.append(self.data['group'][grp][label].loadings.assign(group=grp))
            loadings = pd.concat(chunks, axis='index')
            fig = px.scatter(loadings, x='PC1', y='PC2', color='group',
                             hover_data=loadings.columns.tolist())

        else:
            loadings = self.data[label].loadings.iloc[:, 0:2]
            fig = px.scatter(loadings, x='PC1', y='PC2', hover_data=loadings.columns.tolist())
        fig.update_traces(marker=dict(size=1))

        if standardised:
            fig.add_shape(type="circle",
                          xref="x", yref="y",
                          x0=-1, y0=-1, x1=1, y1=1,
                          line_color="gray")
            title_main = "Correlation Circle"
        else:
            title_main = "Top 2 Principal Components"

        fig = self.add_loading_vectors(fig, loadings)
        title = (f"{title_main}<br>Explained Variance = "
                 f"{round(self.data['std'].explained_variance.iloc[0:2].sum(), 1)}%") if title is None else title

        fig.update_layout(title=title,
                          xaxis_title=f"PC1 ({round(self.data['std'].explained_variance.iloc[0], 1)}%)",
                          yaxis_title=f"PC2 ({round(self.data['std'].explained_variance.iloc[1], 1)}%)")
        fig.update_yaxes(scaleanchor="x", scaleratio=1)
        fig.update_layout(scene=dict(aspectmode="data"))

        return fig

    def add_loading_vectors(self, fig, loadings) -> go.Figure:
        if 'group' in loadings.columns:
            cm = px.colors.qualitative.Plotly
            grp_colors = dict(zip(loadings['group'].unique(), cm[0:len(loadings['group'].unique())]))
        for i, feature in enumerate(loadings.index):
            if 'group' in loadings.columns:
                grp = loadings.iloc[i, :]['group']
                arrowcolor = grp_colors[grp]
                font = dict(color=grp_colors[grp])
            else:
                arrowcolor = None
                font = None
            fig.add_annotation(
                ax=0, ay=0,
                axref="x", ayref="y",
                x=loadings.iloc[i, 0],
                y=loadings.iloc[i, 1],
                showarrow=True,
                arrowsize=2,
                arrowhead=2,
                xanchor="right",
                yanchor="top",
                arrowcolor=arrowcolor,
            )
            fig.add_annotation(
                x=loadings.iloc[i, 0],
                y=loadings.iloc[i, 1],
                ax=0, ay=0,
                xanchor="center",
                yanchor="bottom",
                text=feature,
                yshift=5,
                font=font
            )
        return fig



import numpy as np

from elphick.sklearn_viz.model_selection.metrics import r2_with_nan

classification_scorers = {'acc': 'accuracy',
                          'prec_macro': 'precision_macro',
                          'rec_macro': 'recall_macro'}


def r2_score_with_nan(estimator, X, y):
    return r2_with_nan(y_true=y, y_est=estimator.predict(X))


regression_scorers = {
    # 'r2_nan': r2_score_with_nan,
    'r2_score': 'r2',
    'mae': 'neg_mean_absolute_error',
    'mse': 'neg_mean_squared_error'}



from .importance import FeatureImportance, plot_feature_importance
from .parallel_coordinates import plot_parallel_coordinates
from .principal_components import PrincipalComponents, plot_scatter_matrix, plot_explained_variance, \
    plot_principal_components
from .outlier_detection import OutlierDetection, plot_outlier_matrix



import logging
from typing import Union, Optional, Dict, Callable

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from sklearn.base import is_classifier

from sklearn.inspection import permutation_importance
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from elphick.sklearn_viz.model_selection.scorers import r2_score_with_nan, classification_scorers, regression_scorers
from elphick.sklearn_viz.utils import log_timer


def plot_feature_importance(mdl,
                            sort: bool = False,
                            top_k: Optional[int] = None,
                            horizontal: bool = False,
                            permute: bool = False,
                            pipeline_input_features: bool = False,
                            x_test: Optional[pd.DataFrame] = None,
                            y_test: Optional[Union[pd.DataFrame, pd.Series]] = None,
                            title: Optional[str] = None
                            ) -> go.Figure:
    """

    Args:
        mdl: The scikit-learn model or pipeline.
        sort: If True, sort by decreasing importance
        top_k: Include only the top k features in the plot.  Will ignore the sort argument.
        horizontal: If True plot horizontal bars, if False vertical bars.
        permute: If True plot permutation importance.  Better, but slower.  Requires X_test and y_test to be provided.
        pipeline_input_features: If True, and a pipeline is provided, report the features provided as inputs to
         the pipeline.  If False, reports the estimator (last pipeline step) input features.  Requires permute = True.
        x_test: X values provided to execute permuted importance.
        y_test: y values provided to execute permuted importance.
        title: title for the plot

    Returns:
        a plotly GraphObjects.Figure

    """

    return FeatureImportance(mdl=mdl, permute=permute, pipeline_input_features=pipeline_input_features,
                             x_test=x_test, y_test=y_test).plot(sort=sort, top_k=top_k, horizontal=horizontal,
                                                                title=title)


class FeatureImportance:
    def __init__(self,
                 mdl,
                 permute: bool = False,
                 pipeline_input_features: bool = False,
                 x_test: Optional[pd.DataFrame] = None,
                 y_test: Optional[Union[pd.DataFrame, pd.Series]] = None,
                 scorer: Optional[Union[str, Callable]] = None,):
        """

        Args:
            mdl: The scikit-learn model or pipeline.
            permute: If True plot permutation importance.  Better, but slower.  Requires X_test and y_test to be provided.
            pipeline_input_features: If True, and a pipeline is provided, report the features provided as inputs to
             the pipeline.  If False, reports the estimator (last pipeline step) input features.  Requires permute = True.
            x_test: X values provided to execute permuted importance.
            y_test: y values provided to execute permuted importance.
            scorer: Optional callable scorer which the model will be fitted using

        """
        self._logger = logging.getLogger(name=__class__.__name__)
        self.mdl = mdl
        self.permute: bool = permute
        self.pipeline_input_features: bool = pipeline_input_features
        self.X_test: Optional[pd.DataFrame] = x_test
        self.y_test: Optional[Union[pd.DataFrame, pd.Series]] = y_test

        if scorer is not None:
            self.scorer = scorer
        else:
            self.scorer = classification_scorers[list(classification_scorers.keys())[0]] if is_classifier(self.mdl) else \
                regression_scorers[list(regression_scorers.keys())[0]]

        self._data: Optional[pd.DataFrame] = None
        self.is_pipeline: bool = isinstance(mdl, Pipeline)

        if not self.permute:
            check_is_fitted(mdl[-1]) if self.is_pipeline else check_is_fitted(mdl)

    @property
    @log_timer
    def data(self) -> Optional[pd.DataFrame]:
        if self._data is not None:
            res = self._data
        else:
            mdl = self.mdl

            if self.permute:
                self._logger.info("Generating feature importance by permutation")
                x = self.X_test
                if self.is_pipeline and not self.pipeline_input_features:
                    mdl = mdl[-1]
                    x = self.mdl[0:-1].transform(self.X_test)

                result = permutation_importance(estimator=mdl, X=x, y=self.y_test, n_repeats=10, random_state=42,
                                                n_jobs=2, scoring=self.scorer)
                importances = result.importances_mean
                std = result.importances_std
            else:
                self._logger.info("Extracting feature importance from the fitted model")
                if self.is_pipeline:
                    mdl = mdl[-1]
                try:  # trees
                    importances = mdl.feature_importances_
                    std = np.std([tree.feature_importances_ for tree in mdl.estimators_], axis=0)
                except AttributeError:  # regression
                    importances = mdl.coef_
                    std = np.full(len(importances), np.nan)

            try:
                feature_names = mdl.feature_names_in_
            except AttributeError:
                self._logger.warning("Feature names are not available within the model."
                                     " Setting the transform output to pandas will correct this."
                                     " e.g. pipe.set_output(transform='pandas')."
                                     " Retrying with the pre-processed feature names.")
                try:
                    # Likely non-sklearn estimator like CatBoostRegressor
                    feature_names = list(self.mdl[0:-1].transform(self.X_test).columns)
                except AttributeError:
                    self._logger.warning("Feature names are not available within the model."
                                         " Setting default names")
                    feature_names = [f"F{i}" for i in range(1, mdl.n_features_in_ + 1)]

            res: pd.DataFrame = pd.DataFrame([importances, std], index=['importance', 'std'], columns=feature_names).T
            self._data = res

        return res

    def plot(self,
             sort: bool = False,
             top_k: Optional[int] = None,
             horizontal: bool = False,
             title: Optional[str] = None) -> go.Figure:
        """

        Args:
            sort: If True, sort by decreasing importance
            top_k: Include only the top k features in the plot.  Will ignore the sort argument.
            horizontal: If True plot horizontal bars, if False vertical bars.
             the pipeline.  If False, reports the estimator (last pipeline step) input features.  Requires permute = True.
            title: title for the plot

        Returns:
            a plotly GraphObjects.Figure

        """
        data = self.data
        subtitle: str = 'Feature Importance'
        if self.permute:
            subtitle = 'Permuted ' + subtitle

        if sort or top_k is not None:
            data = data.sort_values(by=['importance'], ascending=False)
            if top_k is not None:
                data = data.iloc[0:top_k, :]
            if horizontal:
                data = data.sort_values(by=['importance'], ascending=True)

        if horizontal:
            kwargs: Dict = {'y': data.index,
                            'x': data['importance'],
                            'error_x': dict(type='data', array=data['std']),
                            'orientation': 'h'}
        else:
            kwargs: Dict = {'x': data.index,
                            'y': data['importance'],
                            'error_y': dict(type='data', array=data['std'])}

        if title is None:
            title = subtitle
        else:
            title = title + '<br>' + subtitle

        fig = go.Figure()
        fig.add_trace(go.Bar(name='Model 1', **kwargs))
        fig.update_layout(title=title, barmode='group')

        return fig



from sklearn.model_selection import StratifiedKFold


class FeatureStratifiedKFold:
    """Strategy to split data into folds based on a feature column.

    The standard StratifiedKFold class from scikit-learn supports stratification based on a feature column.
    """

    def __init__(self, n_splits=5, shuffle=True, random_state=42):
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state
        self.skfold = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)

    def split(self, X, y=None, groups=None):
        if groups is None:
            raise ValueError("The 'groups' parameter should not be None")
        return self.skfold.split(X, groups)

    def get_n_splits(self, X, y, groups=None):
        return self.n_splits



from dataclasses import dataclass
from enum import Enum
from typing import Union, List, Dict

import sklearn.base
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.linear_model import LogisticRegression, LinearRegression, LassoCV, RidgeCV, ElasticNetCV
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


class ModelType(Enum):
    REGRESSOR = 'regressor'
    CLASSIFIER = 'classifier'


@dataclass
class Model:
    code: str
    model: Union[sklearn.base.RegressorMixin, sklearn.base.ClassifierMixin]
    type: ModelType
    fast: bool


class Models:
    def __init__(self):
        self.LR_C = Model('LR', LogisticRegression(), ModelType.CLASSIFIER, True)
        self.LDA = Model('LDA', LinearDiscriminantAnalysis(), ModelType.CLASSIFIER, True)
        self.KNN = Model('KNN', KNeighborsClassifier(), ModelType.CLASSIFIER, True)
        self.CART_C = Model('CART', DecisionTreeClassifier(), ModelType.CLASSIFIER, True)
        self.NB = Model('NB', GaussianNB(), ModelType.CLASSIFIER, True)
        self.SVM = Model('SVM', SVC(), ModelType.CLASSIFIER, True)
        self.RF_C = Model('RF', RandomForestClassifier(), ModelType.REGRESSOR, False)
        self.LR_R = Model('LR', LinearRegression(), ModelType.REGRESSOR, True)
        self.LASSO = Model('LASSO', LassoCV(), ModelType.REGRESSOR, True)
        self.RIDGE = Model('RIDGE', RidgeCV(), ModelType.REGRESSOR, True)
        self.ELASTIC = Model('ELASTIC', ElasticNetCV(), ModelType.REGRESSOR, True)
        self.CART_R = Model('CART', DecisionTreeRegressor(), ModelType.REGRESSOR, True)
        self.RF_R = Model('RF', RandomForestRegressor(), ModelType.REGRESSOR, False)

    def fast_classifiers(self) -> Dict:
        return {v.code: v.model for k, v in self.__dict__.items() if
                v.type == ModelType.CLASSIFIER and v.fast is True}

    def all_classifiers(self) -> Dict:
        return {v.code: v.model for k, v in self.__dict__.items() if v.type == ModelType.CLASSIFIER}

    def fast_regressors(self) -> Dict:
        return {v.code: v.model for k, v in self.__dict__.items() if v.type == ModelType.REGRESSOR and v.fast is True}

    def all_regressors(self) -> Dict:
        return {v.code: v.model for k, v in self.__dict__.items() if v.type == ModelType.REGRESSOR}



from functools import partial
from typing import Literal

import numpy as np
from sklearn.metrics import r2_score, mean_absolute_error, mean_squared_error, f1_score


def mean_error(y_true, y_est) -> float:
    return float(np.mean(y_est - y_true))


def moe_95(y_true, y_est, metric: Literal["moe", "lo", "hi"] = 'moe') -> float:
    me = mean_error(y_true, y_est)
    rmse = mean_squared_error(y_true, y_est, squared=False)
    if metric == 'lo':
        res = me - (rmse * 1.96)
    elif metric == 'hi':
        res = me + (rmse * 1.96)
    elif metric == 'moe':
        res = np.mean([(np.abs(me - (rmse * 1.96))), (me + (rmse * 1.96))])
    else:
        raise KeyError(f'Invalid metric supplied.  Allowed values are: {Literal["moe", "lo", "hi"]}')
    return res


def r2(y_true, y_est):
    ssr = ((y_true - y_est) ** 2).sum()
    sst = ((y_true - y_true.mean()) ** 2).sum()
    return 1 - (ssr / sst)


def r2_with_nan(y_true, y_est):
    numerator = np.nansum((y_true - y_est) ** 2, axis=0, dtype=np.float64)
    denominator = np.nansum((y_true - np.nanmean(y_true, axis=0)) ** 2, axis=0, dtype=np.float64)
    return np.mean(1 - numerator / denominator)


regression_metrics = {
    # 'r2_nan': r2_with_nan,
    'r2_score': r2_score,
    'me': mean_error,
    'r2': r2,
    'mae': mean_absolute_error,
    'mse': mean_squared_error,
    'rmse': partial(mean_squared_error, squared=False),
    'moe': partial(moe_95, metric='moe'),
    'moe_lo': partial(moe_95, metric='lo'),
    'moe_hi': partial(moe_95, metric='hi')}

classification_metrics = {'f1': f1_score}
